package rbac

import (
	"encoding/json"
	"errors"

	"code.fbi.h-da.de/danet/gosdn/controller/conflict"
	"code.fbi.h-da.de/danet/gosdn/controller/interfaces/rbac"
	"github.com/google/uuid"
	"go.mongodb.org/mongo-driver/bson"
)

// User represents the data of a user for access control and is stored in a storage.
type User struct {
	UserID   uuid.UUID         `json:"_id"`
	Metadata conflict.Metadata `json:"metadata"`
	UserName string            `json:"username"`
	Roles    map[string]string `json:"roles,omitempty"`
	Password string            `json:"password"`
	Tokens   []string          `json:"token,omitempty"`
	Salt     string            `json:"salt"`
}

// NewUser creates a new user.
func NewUser(id uuid.UUID,
	name string,
	roles map[string]string,
	pw string,
	tokens []string,
	salt string,
	metadata conflict.Metadata,
) rbac.User {
	return &User{
		UserID:   id,
		UserName: name,
		Roles:    roles,
		Password: pw,
		Tokens:   tokens,
		Salt:     salt,
		Metadata: metadata,
	}
}

// ID returns a UUID of the user.
func (u *User) ID() uuid.UUID {
	return u.UserID
}

// GetMetadata returns the metadata of a user.
func (u *User) GetMetadata() conflict.Metadata {
	return u.Metadata
}

// Name returns the name of the user.
func (u *User) Name() string {
	return u.UserName
}

// GetPassword returns the password of the user.
func (u *User) GetPassword() string {
	return u.Password
}

// GetRoles returns the roles of the user.
func (u *User) GetRoles() map[string]string {
	return u.Roles
}

// GetToken returns the token of the user.
func (u *User) GetTokens() []string {
	return u.Tokens
}

// SetName sets the name of the user.
func (u *User) SetName(name string) {
	u.UserName = name
}

// SetToken sets the token of the user.
func (u *User) SetTokens(tokens []string) {
	u.Tokens = tokens
}

// AddToken adds a token to the user.
func (u *User) AddToken(token string) {
	u.Tokens = append(u.Tokens, token)
}

// RemoveToken removes a token from the user.
func (u *User) RemoveToken(inputToken string) error {
	for i, token := range u.Tokens {
		if token == inputToken {
			u.Tokens = append(u.Tokens[:i], u.Tokens[i+1:]...)
			return nil
		}
	}
	err := errors.New("token not found")
	return err
}

// GetSalt returns the salt of the user.
func (u *User) GetSalt() string {
	return u.Salt
}

// MarshalJSON implements the MarshalJSON interface to store a user as JSON.
func (u *User) MarshalJSON() ([]byte, error) {
	return json.Marshal(&struct {
		UserID   uuid.UUID         `json:"_id"`
		UserName string            `json:"username"`
		Roles    map[string]string `json:"roles,omitempty"`
		Password string            `json:"password"`
		Tokens   []string          `json:"tokens,omitempty"`
		Salt     string            `json:"salt"`
		Metadata conflict.Metadata `json:"metadata"`
	}{
		UserID:   u.ID(),
		UserName: u.Name(),
		Roles:    u.Roles,
		Password: u.Password,
		Tokens:   u.Tokens,
		Salt:     u.Salt,
		Metadata: u.Metadata,
	})
}

// MarshalBSON implments the MarshalBSON interface to store a user as BSON.
func (u *User) MarshalBSON() ([]byte, error) {
	return bson.Marshal(&struct {
		UserID   string            `bson:"_id"`
		UserName string            `bson:"username"`
		Roles    map[string]string `bson:"roles,omitempty"`
		Password string            `bson:"password"`
		Tokens   []string          `bson:"tokens,omitempty"`
		Salt     string            `bson:"salt"`
		Metadata conflict.Metadata `bson:"metadata"`
	}{
		UserID:   u.ID().String(),
		UserName: u.Name(),
		Roles:    u.Roles,
		Password: u.Password,
		Tokens:   u.Tokens,
		Salt:     u.Salt,
		Metadata: u.Metadata,
	})
}
